.text
# Loads the OS from disk block 1 on into location zero of memory.
# Since we're living at location 0, we need to move ourselves out of
# the way first.
__start:
	addi	$t0, $0, 0x4000		# form mem size address
	addi	$t1, $0, 16
	sllv	$t0, $t0, $t1
	lw	$t0, 0x28($t0)		# get memory size
	addi	$t1, $t0, -512		# back off a ways
	or	$t5, $0, $t1		# save base address of copy
	add	$t2, $0, $0		# next instruction to copy
	# here's the loop to copy ourself
copy:	lw	$t3, 0($t2)
	sw	$t3, 0($t1)
	addi	$t1, $t1, 4
	addi	$t2, $t2, 4
	slt	$t4, $t1, $t0
	bne	$t4, $0, copy
	addi	$t5, $t5, resume	# compute address of resume in copy
	jr	$t5			# and go...

resume:	
	# now go ahead and load the OS
	addi	$t0, $0, 1		# address of block device
	addi	$at, $0, 30
	sllv	$t0, $t0, $at
	addi	$t0, $t0, 16		# block device address
	addi	$t2, $0, 1		# block to load
	ori	$t3, $0, 0x8000		# command (read)
	sw	$t2, 12($t0)		# read 1 block
	sw	$t2, 8($t0)		# block 1 on the disk
	sw	$0, 4($t0)		# address 0
	sw	$t3, 0($t0)		# go
	ori	$at, $0, 0x2000		# done mask
wait:	lw	$t4, 0($t0)		# wait until it finishes
	xor	$t4, $t4, $at
	bne	$t4, $0, wait
	sw	$at, 0($t0)		# clear done flag in block device
	
	# go ahead and start reading at 4, just skip the magic number!
	lw	$gp, 4($0)		# text size (global pointer)
	lw	$t5, 8($0)		# data size
	lw	$t6, 12($0)		# stack size
	lw	$t7, 16($0)		# heap size
	lw	$t8, 20($0)		# entry point
	add	$at, $gp, $t5
	add	$at, $at, $t6
	add	$sp, $at, $t7		# set up the stack pointer
	addi	$sp, $sp, -4
	
	add	$t5, $gp, $t5		# t5 holds total OS text+data size
	addi	$t5, $t5, 24		# increment by size of executable header (24 bytes)
	addi	$t5, $t5, 511		# start computing total blocks in os image
	addi	$t6, $0, 512		# block size
	div	$t5, $t5, $t6		# number of blocks to in image
	addi	$t5, $t5, -1		# number of blocks to read
	sw	$t5, 12($t0)		# set block count in block device
	sw	$t6, 4($t0)		# set memory address (512)
	ori	$t2, $0, 2		# want to start reading from disk at block 2
	sw	$t2, 8($t0)
	sw	$t3, 0($t0)		# read all the blocks
	# if the disk file length isn't a multiple of 512, an exception (#2) will occur
	ori	$at, $0, 0x2000
wait2:	lw	$t4, 0($t0)		# wait until it finishes
	and	$t4, $t4, $at
	beq	$t4, $0, wait2
	sw	$at, 0($t0)		# clear exception
	
done:	addi	$t6, $sp, 4		# end of memory claimed by OS
	addi	$t5, $t5, 2		# disk block number of first executable
	addi	$sp, $sp, -8		# prepare to push arg to os init()
	sw	$t6, 0($sp)		# push os memory size
	sw	$t5, 4($sp)		# push first executable block number
	jr	$t8                     # jump to the entry point of the OS

