int charDev =  0x40000000;
int blockDev = 0x40000010;
int memDev =   0x40000020;
int trapDev =  0x40000030;

int maxProcs = 4;
int currentProc = -1;  // index of current executing process

int memHighWater = 0; // memory above this address is available

int isRoot[4];        // is process n running as root?
int procBase[4];      // base address of each process
int procLength[4];    // amount of memory allocated to each process
int savedPC[4];       // process PC
int proc0State[32];   // proc0 saved regs
int proc1State[32];   // proc1 saved regs
int proc2State[32];   // proc2 saved regs
int proc3State[32];   // proc3 saved regs

int GP = 28;
int SP = 29;
int FP = 30;
int RA = 31;

// this is the entry called by the bootloader.
// set the memory high water and then launch the shell process.
init(int* pHandler, int osMemSize, int shellBlock) {
  int* pTrapDev = trapDev;
  int* pCharDev = charDev;

  // initialize trap handler address
  pTrapDev[0] = pHandler;

  // initialize mem high water
  //    memHighWater = (osSize+3) & 0xfffffffc;  
  // doesn't work because ceb doesn't like the literal
  memHighWater = ((osMemSize+3) | 3) ^ 3;

  _printString(pCharDev, "init(): shellBlock = ");
  _printInt(pCharDev, shellBlock);
  _printString(pCharDev, "\n");

  // start the shell
  syscall(11, shellBlock, 0 );
}

dispatch() {
  int*   pCharDev = charDev;
  int*   pMemDev = memDev;
  int*   pTrapDev = trapDev;
  int*   savedRegs;

  pMemDev[0] = procBase[currentProc];
  pMemDev[1] = procLength[currentProc];
  pTrapDev[2] = savedPC[currentProc];

  //_printString(pCharDev, "dispatch(): membase = ");
  //_printInt(pCharDev, pMemDev[0]);
  //_printString(pCharDev, " memlen = ");
  //_printInt(pCharDev, pMemDev[1]);
  //_printString(pCharDev, " PC = ");
  //_printInt(pCharDev, pTrapDev[2]);
  //_printString(pCharDev, "\n");

  if ( currentProc == 0 ) {
    savedRegs = proc0State;
  }
  if ( currentProc == 1 ) {
    savedRegs = proc1State;
  }
  if ( currentProc == 2 ) {
    savedRegs = proc2State;
  }
  if ( currentProc == 3 ) {
    savedRegs = proc3State;
  }
  if ( (currentProc < 0) | (currentProc > 3) != 0) {
    _printString(pCharDev, "Kernel crash: dispatch() with currentProc = ");
    _printInt(pCharDev, currentProc);
    _printString(pCharDev, "\n");
    savedRegs = proc0State;
  }
  __dispatch(savedRegs);
}

// when we get to this routine, if it's a syscall the 
// arg(s) are in the saved registers
exceptionHandler() {
  int*  pCharDev = charDev;
  int*  pTrapDev = trapDev;
  int*  pCurrentProcSavedRegs;
  int   cause;

  savedPC[currentProc] = pTrapDev[2];
  cause = pTrapDev[3];

  pCurrentProcSavedRegs = 0;   // address 0
  pCurrentProcSavedRegs = pCurrentProcSavedRegs[2];  // Mem[8] -- pointer to saved regs

  //_printString(pCharDev, "exceptionHandler(): cause = ");
  //_printInt(pCharDev, cause);
  //_printString(pCharDev, " savedPC = ");
  //_printInt(pCharDev, savedPC[currentProc]);
  //_printString(pCharDev, " currentProc = ");
  //_printInt(pCharDev, currentProc);
  //_printString(pCharDev, "\n");

  if ( cause == 16 ) {    // syscall
    savedPC[currentProc] = savedPC[currentProc]+4;  // don't want to do syscall again
    int type = pCurrentProcSavedRegs[2];  // register $v0
    int arg1 = pCurrentProcSavedRegs[4];  // register $a0
    int arg2 = pCurrentProcSavedRegs[5];  // register $a1
    syscall(type, arg1, arg2); // never returns
  }
  else {
    if ( cause == 12 ) { // overflow
      _printString(pCharDev, "Integer overflow exception.\n");
      currentProc = currentProc - 1;
      dispatch();
    }
    else  {
      _printString(pCharDev, "Unhandled exception type: ");
      _printInt(pCharDev, cause);
      _printString(pCharDev, "\n");
      dispatch();
    }
  }
}

// convert the address the user gave into a real address
int translateAddress(int virtual) {
  int* pMemDev = memDev;
  int  base = pMemDev[0];
  int  length = pMemDev[1];

  if (virtual > length) {
    _printString(charDev, "Addressing error on syscall");
    syscall( 10, 0, 0 ); // exit process
  }
  return virtual+base;
}

// handle a syscall
syscall(int type, int* arg1, int* arg2) {
  int  physicalAddress;
  int  exeData[5];
  int* term = charDev;
  int* pBlockDev = blockDev;
  int  result = 0;

  //_printString(term, "In syscall, type=");
  //_printInt(term, type);
  //_printString(term, " arg=");
  //_printInt(term, arg1);

  if (type == 1) {
    _printInt(term, arg1);
  }
  if (type == 4) {
    physicalAddress = translateAddress(arg1);
    _printString(term, physicalAddress);
  }
  if (type == 8) {
    physicalAddress = translateAddress(arg1);
    _readString(term, physicalAddress, arg2);  // security breach here (if arg2 is large)
  }
  if (type == 5) {
    int* pCurrentProcSavedRegs;
    pCurrentProcSavedRegs = 0;   // address 0
    pCurrentProcSavedRegs = pCurrentProcSavedRegs[2];  // Mem[8] -- pointer to saved regs
    pCurrentProcSavedRegs[2] = _readInt(term);  // put result in current proc's $v0
  }
  if (type == 11) {
    // exec call
      if ( currentProc < maxProcs-1 ) {

	currentProc = currentProc + 1;

	_printString(term, "\nSyscall exec: trying to start process ");
	_printInt(term, currentProc);
	_printString(term, "\n");

	_load(pBlockDev, arg1, memHighWater, exeData);

	int exeSize = exeData[0] + exeData[1] + exeData[2] + exeData[3];

	procBase[currentProc] = memHighWater;
	procLength[currentProc] = exeSize;
	savedPC[currentProc] = exeData[4];
	
	_printString(term, "Proc base/length/entry = ");
	_printInt(term, procBase[currentProc]);
	_printString(term, " / ");
	_printInt(term, procLength[currentProc]);
	_printString(term, " / ");
	_printInt(term, savedPC[currentProc]);
	_printString(term, "\n");

	if ( currentProc== 0 ) {
	  proc0State[GP] = exeData[0];
	  proc0State[SP] = exeSize - 4;
	  proc0State[FP] = 0;   // ?? copied from original code
	}
	if ( currentProc== 1 ) {
	  proc1State[GP] = exeData[0];
	  proc1State[SP] = exeSize - 4;
	  proc1State[FP] = 0;   // ?? copied from original code
	}
	if ( currentProc== 2 ) {
	  proc2State[GP] = exeData[0];
	  proc2State[SP] = exeSize - 4;
	  proc2State[FP] = 0;   // ?? copied from original code
	}
	if ( currentProc== 3 ) {
	  proc3State[GP] = exeData[0];
	  proc3State[SP] = exeSize - 4;
	  proc3State[FP] = 0;   // ?? copied from original code
	}

	memHighWater = memHighWater + exeSize;
	memHighWater = ((memHighWater+3) | 3) ^ 3;  // round up to word address

	_printString(term, "Process size: ");
	_printInt(term, exeSize);
	_printString(term, "\n");
      }
      else {
	_printString(charDev, "Process limit exceeded");
	type = 10;  // kill the current process by flowing into exit syscall code
      }
  }
  if (type == 10) {
    //_printString(term, "Exit syscall");
    currentProc = currentProc - 1;
  }

  dispatch();
}

// Wait until the device has an input character
_waitForCharInput(int* p) {
  while ((p[0] & 0x4000) == 0) {}
}

// Wait until the device is ready for output
_waitForCharOutput(int* p) {
  while ((p[0] & 0x8000) == 0) {}
}

_readString(int* dev, char* buf, int len) {
  int i, ch;
  i=0;
  len = len - 1;
  while (i < len) {
    _waitForCharInput(dev);
    ch = dev[0] & 0xff;  // fetch the character
    dev[0] = 0x4000;     // clear the input char ready flag
    if (ch == 13) {
      buf[i] = 0;
      return 1;
    }
    else {
      if (ch != 10) {
	buf[i] = ch;
      }
      else {
	buf[i] = 0;
	return 1;
      }
    }
    i = i + 1;
  }

  return 0;
}

_readInt(int* dev) {
  char buffer[80];
  int negative = 0;
  int i = 0;
  int result = 0;

  _readString(dev, buffer, 80);

  if (buffer[0] == 45) {
    negative = 1;
    i = 1;
  }
  while (buffer[i] != 0) {
    result = (result * 10) + (buffer[i] - 48);
    i = i + 1;
  }

  if (negative) {
    return -result;
  }
  else {
    return result;
  }
}

_strlen(char* s) {
  int i=0;
  while (s[i] != 0) {
    i = i + 1;
  }
  return i;
}

// my very own print string, using memory mapped IO
_printString(int* dev, char* buf) {
  int i;
  i=0;
  while (buf[i] != 0) {
    _waitForCharOutput(dev);
    dev[0] = 0x00008000 + buf[i];
    i = i + 1;
  }
}

_printInt(char* dev, int val) {
  char buffer1[20];
  char buffer2[20];
  int i, j;
  j=0;
  if (val < 0) {
    val = -val;
    buffer2[0] = 45;    // - sign
    j = 1;
  }
  buffer1[0] = (val % 10) + 48;
  i = 1;
  val = val / 10;
  while (val != 0) {
    buffer1[i] = (val % 10) + 48;
    val = val / 10;
    i = i + 1;
  }
  i=i-1;
  while (i >= 0) {
    buffer2[j] = buffer1[i];
    j = j + 1;
    i = i - 1;
  }
  buffer2[j] = 0;
  _printString(dev, buffer2);
}

_readBlock(int* dev, char* buffer, int block) {
  int  pCharDev = charDev;
  int  exceptionCode;

  while ( (dev[0]& 0xe000) != 0 ) {}    // wait for device not busy
  dev[1] = buffer;
  dev[2] = block;
  dev[3] = 1;
  dev[0] = 0x8000;
  while ( (dev[0] & 0x2000) == 0 ) {}
  exceptionCode = dev[0] & 0xff;
  if ( exceptionCode != 0 ) {
    _printString(pCharDev, "Block device exception: ");
    _printInt(pCharDev, exceptionCode);
    _printString(pCharDev, "\n");
  }
  dev[0] = 0x2000;   // clear exception/done on block controller
}

_copy(int* dest, int* src, int words) {
  int i;
  i=0;
  while (i<words) {
    dest[i] = src[i];
    i = i + 1;
  }
}

// load the executable that lives at the given block on
// disk to the given address in memory.
// fills in the exeData structure with an executable description
_load(int* dev, int block, int address, int* exeData) {
  int buffer[128];  // 512 bytes
  int bytesCopied, size;
  int term = charDev;

  //_printString(term, "Loading exe at block ");
  //_printInt(term, block);
  //_printString(term, " to mem at address ");
  //_printInt(term, address);
  //_printString(term, "\n");

  _readBlock(dev, buffer, block);

  exeData[0]=buffer[1];   // text size
  exeData[1]=buffer[2];   // data size
  exeData[2]=buffer[3];   // stack size
  exeData[3]=buffer[4];   // heap size
  exeData[4]=buffer[5];   // entry point

  //_printString(term, "Loading process\ntext size = ");
  //_printInt(term, exeData[0]);
  //_printString(term, "\ndata size = "); 
  //_printInt(term, exeData[1]);
  //_printString(term, "\nstack size ="); 
  //_printInt(term, exeData[2]);
  //_printString(term, "\nheap size ="); 
  //_printInt(term, exeData[3]);
  //_printString(term, "\nentry = "); 
  //_printInt(term, exeData[4]);
  //_printString(term, "\n");

  _copy(address, buffer + 24, 122);
  bytesCopied = 488;
  size = (exeData[0] + exeData[1]);
  block = block + 1;
  while (bytesCopied < size) {
    _readBlock(dev, buffer, block);
    _copy(address+bytesCopied, buffer, 128);
    bytesCopied = bytesCopied + 512;
    block = block + 1;
  }

  //_printString(term, "Load complete.\n");
}





